/* $Id: util.c 533 2007-12-22 15:25:42Z lukasz $ */
/*
 *  FedStage DRMAA for PBS Pro
 *  Copyright (C) 2006-2007  Fedstage Systems Inc.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * @file pbs_drmaa/util.c
 * PBS DRMAA utilities.
 */

#ifdef HAVE_CONFIG_H
#	include <config.h>
#endif

#include <stdlib.h>
#include <string.h>

#include <pbs_error.h>

#include <drmaa_utils/drmaa_base.h>
#include <pbs_drmaa/pbs_drmaa.h>


void
drmaa_log_attrl( const struct attrl *attribute_list, const char *prefix )
{
	const struct attrl *i;

	if( prefix == NULL )
		prefix = "";
	for( i = attribute_list;  i != NULL;  i = i->next )
		drmaa_log( " %s %s%s%s=%s",
				prefix, i->name,
				i->resource ? "." : "",  i->resource ? i->resource : "",
				i->value
				);
}


void
drmaa_free_attrl( struct attrl *attr )
{
	while( attr != NULL )
	 {
		struct attrl *p = attr;
		attr = attr->next;
		free( p->name     );
		free( p->value    );
		free( p->resource );
		free( p           );
	 }
}


struct pbs_attrib { const char *name; int code; };

const struct pbs_attrib *
pbs_attrib_lookup( const char *str, unsigned int len );

const drmaa_attrib_info_t *
attr_by_drm_name( drmaa_session_t *session, const char *drm_name,
	drmaa_err_ctx_t *err )
{
	const struct pbs_attrib *attr = NULL;
	attr = pbs_attrib_lookup( drm_name, strlen(drm_name) );
	if( attr == NULL )
	 {
		/* RAISE_DRMAA_1( DRMAA_ERRNO_INVALID_ARGUMENT ); */
		return NULL;
	 }
	else if( attr->code < session->n_attributes )
		return & session->attributes[ attr->code ];
	else
	 {
		RAISE_DRMAA_1( DRMAA_ERRNO_INTERNAL_ERROR );
		return NULL;
	 }
}


void
drmaa_err_pbs_error( drmaa_err_ctx_t *err )
{
	const char *msg = NULL;
	DEBUG(( "-> drmaa_err_pbs_error" ));
	if( ++err->n_errors > 1 )
		return;
	/*
	 * Gathering error messages differ between PBS forks.
	 * - OpenPBS - ...
	 * - Torque - pbse_to_txt takes PBS error code (stored in pbs_errno)
	 *  and returns corresponding error message.
	 * - PBS Pro - stores errno of last operation inside pbs_errno variable;
	 *  pbse_to_txt always return NULL.
	 * All of them define pbs_geterrmsg which returns last error message
	 * for given connection.
	 */
	/* XXX: PBSPro has some link problems with pbse_to_txt function */
#if 0
	msg = pbse_to_txt( pbs_errno );
#else
	msg = "pbs error";
#endif
	if( err->msg != NULL )
		snprintf( err->msg, err->msgsize, "pbs-drm: %s", msg );
	err->rc = drmaa_err_map_pbs_error( pbs_errno );
	LOG_STACKTRACE();
	DEBUG(( "<- drmaa_err_pbs_error; pbs_errno=%d: %s; err->rc=%d",
				pbs_errno, msg, err->rc ));
}


/** Maps PBS error code into DMRAA code. */
int
drmaa_err_map_pbs_error( int pbs_errcode )
{
	DEBUG(( "-> drmaa_err_map_pbs_error( pbs_errcode=%d )", pbs_errcode ));
	switch( pbs_errcode )
	 {
		case PBSE_NONE:  /* no error */
			return DRMAA_ERRNO_SUCCESS;
		case PBSE_UNKJOBID:	 /* Unknown Job Identifier */
			return DRMAA_ERRNO_INVALID_JOB;
		case PBSE_NOATTR: /* Undefined Attribute */
		case PBSE_ATTRRO: /* attempt to set READ ONLY attribute */
		case PBSE_IVALREQ:  /* Invalid request */
		case PBSE_UNKREQ:  /* Unknown batch request */
			return DRMAA_ERRNO_INTERNAL_ERROR;
		case PBSE_PERM:  /* No permission */
		case PBSE_BADHOST:  /* access from host not allowed */
			return DRMAA_ERRNO_AUTH_FAILURE;
		case PBSE_JOBEXIST:  /* job already exists */
		case PBSE_SVRDOWN:  /* req rejected -server shutting down */
		case PBSE_EXECTHERE:  /* cannot execute there */
		case PBSE_NOSUP:  /* Feature/function not supported */
		case PBSE_EXCQRESC:  /* Job exceeds Queue resource limits */
		case PBSE_QUENODFLT:  /* No Default Queue Defined */
		case PBSE_NOTSNODE:  /* no time-shared nodes */
			return DRMAA_ERRNO_DENIED_BY_DRM;
		case PBSE_SYSTEM:  /* system error occurred */
		case PBSE_INTERNAL:  /* internal server error occurred */
		case PBSE_REGROUTE:  /* parent job of dependent in rte que */
		case PBSE_UNKSIG:  /* unknown signal name */
			return DRMAA_ERRNO_INTERNAL_ERROR;
		case PBSE_BADATVAL:  /* bad attribute value */
		case PBSE_BADATLST:  /* Bad attribute list structure */
		case PBSE_BADUSER:  /* Bad user - no password entry */
		case PBSE_BADGRP:  /* Bad Group specified */
		case PBSE_BADACCT:  /* Bad Account attribute value */
		case PBSE_UNKQUE:  /* Unknown queue name */
		case PBSE_UNKRESC:  /* Unknown resource */
		case PBSE_UNKNODEATR:  /* node-attribute not recognized */
		case PBSE_BADNDATVAL:  /* Bad node-attribute value */
		case PBSE_BADDEPEND:  /* Invalid dependency */
		case PBSE_DUPLIST:  /* Duplicate entry in List */
			return DRMAA_ERRNO_INVALID_ATTRIBUTE_VALUE;
		case PBSE_MODATRRUN:  /* Cannot modify attrib in run state */
		case PBSE_BADSTATE:  /* request invalid for job state */
		case PBSE_BADCRED:  /* Invalid Credential in request */
		case PBSE_EXPIRED:  /* Expired Credential in request */
		case PBSE_QUNOENB:  /* Queue not enabled */
			return DRMAA_ERRNO_INTERNAL_ERROR;
		case PBSE_QACESS:  /* No access permission for queue */
			return DRMAA_ERRNO_AUTH_FAILURE;
		case PBSE_HOPCOUNT:  /* Max hop count exceeded */
		case PBSE_QUEEXIST:  /* Queue already exists */
		case PBSE_ATTRTYPE:  /* incompatable queue attribute type */
			return DRMAA_ERRNO_INTERNAL_ERROR;
#		ifdef PBSE_QUEBUSY
		case PBSE_QUEBUSY:  /* Queue Busy (not empty) */
#		endif
		case PBSE_MAXQUED:  /* Max number of jobs in queue */
		case PBSE_NOCONNECTS:  /* No free connections */
		case PBSE_TOOMANY:  /* Too many submit retries */
		case PBSE_RESCUNAV:  /* Resources temporarily unavailable */
			return DRMAA_ERRNO_TRY_LATER;
		case 111:
		case PBSE_PROTOCOL:  /* Protocol (ASN.1) error */
		case PBSE_DISPROTO:  /* Bad DIS based Request Protocol */
			return DRMAA_ERRNO_DRM_COMMUNICATION_FAILURE;
#if 0
		case PBSE_QUENBIG:  /* Queue name too long */
		case PBSE_QUENOEN:  /* Cannot enable queue,needs add def */
		case PBSE_NOSERVER:  /* No server to connect to */
		case PBSE_NORERUN:  /* Job Not Rerunnable */
		case PBSE_ROUTEREJ:  /* Route rejected by all destinations */
		case PBSE_ROUTEEXPD:  /* Time in Route Queue Expired */
		case PBSE_MOMREJECT:  /* Request to MOM failed */
		case PBSE_BADSCRIPT:  /* (qsub) cannot access script file */
		case PBSE_STAGEIN:  /* Stage In of files failed */
		case PBSE_CKPBSY:  /* Checkpoint Busy, may be retries */
		case PBSE_EXLIMIT:  /* Limit exceeds allowable */
		case PBSE_ALRDYEXIT:  /* Job already in exit state */
		case PBSE_NOCOPYFILE:  /* Job files not copied */
		case PBSE_CLEANEDOUT:  /* unknown job id after clean init */
		case PBSE_NOSYNCMSTR:  /* No Master in Sync Set */
		case PBSE_SISREJECT:  /* sister rejected */
		case PBSE_SISCOMM:  /* sister could not communicate */
		case PBSE_CKPSHORT:  /* not all tasks could checkpoint */
		case PBSE_UNKNODE:  /* Named node is not in the list */
		case PBSE_NONODES:  /* Server has no node list */
		case PBSE_NODENBIG:  /* Node name is too big */
		case PBSE_NODEEXIST:  /* Node name already exists */
		case PBSE_MUTUALEX:  /* State values are mutually exclusive */
		case PBSE_GMODERR:  /* Error(s) during global modification of nodes */
		case PBSE_NORELYMOM:  /* could not contact Mom */
			return DRMAA_ERRNO_INTERNAL_ERROR;
#endif
		default:
			return DRMAA_ERRNO_INTERNAL_ERROR;
	 }
}


