# -*- coding: UTF-8 -*-
#
# $Id: drmaa_testsuite.py 533 2007-12-22 15:25:42Z lukasz $
#
# Testsuite for DRMAA library
# Copyright (C) 2006-2007  Fedstage Systems Inc.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import unittest
import os.path as osp
import subprocess
import random, time

import drmaa
from drmaa.testsuite.mt_utils import *
from drmaa.testsuite.utils import *
import testing.debug


def testsuite():
	suite = unittest.TestSuite()
	suite.addTests( [
			Test_session_opening("test_multi_init"),
			Test_no_session("test_control_terminate"),
			Test_no_session("test_job_ps"),
			Test_no_session("test_wait"),
			Test_no_session("test_synchronize"),
			Test_no_session("test_exit"),
			Test_single_threaded("test_empty_session_wait"),
			Test_single_threaded("test_submit_wait"),
			Test_single_threaded("test_submit_in_hold"),
			Test_multi_threaded("test_submit_terminate_wait"),
			Test_multi_threaded("test_control_jobs"),
			Test_multi_threaded("test_stress"),
			] )
	suite.addTests(
			unittest.TestLoader().loadTestsFromTestCase( DRMAA_C_testsuite ) )
	if any( [drm.find("PBS") >= 0 for drm in drmaa.DRM_system() ] ):
		suite.addTests( [] )
	return suite
	


class Test_session_opening( unittest.TestCase ):
	""" Test unusual session opening/closing """
	def tearDown( self ):
		import gc
		gc.collect()

	def test_multi_init( self ):
		first = drmaa.Session()
		drmaa_xfail( lambda: drmaa.Session(),
				drmaa.ERRNO_ALREADY_ACTIVE_SESSION )
		first.exit()
		del first


class Test_no_session( unittest.TestCase ):
	""" Test behaviour of DRMAA functions after session is closed """
	def setUp( self ):
		self.session = drmaa.Session()
		drmaa._exit_session()

	def tearDown( self ):
		del self.session

	@drmaa_xfail( drmaa.ERRNO_NO_ACTIVE_SESSION )
	def test_control_terminate( self ):
		self.session.control( "foo", drmaa.TERMINATE )

	@drmaa_xfail( drmaa.ERRNO_NO_ACTIVE_SESSION )
	def test_job_ps( self ):
		self.session.job_ps( "foo" )

	@drmaa_xfail( drmaa.ERRNO_NO_ACTIVE_SESSION )
	def test_wait( self ):
		self.session.wait( "foo" )

	@drmaa_xfail( drmaa.ERRNO_NO_ACTIVE_SESSION )
	def test_synchronize( self ):
		self.session.synchronize( ["foo"] )

	@drmaa_xfail( drmaa.ERRNO_NO_ACTIVE_SESSION )
	def test_exit( self ):
		drmaa._exit_session()



class Test_single_threaded( unittest.TestCase ):
	def setUp( self ):
		self.session = drmaa.Session()

	def tearDown( self ):
		self.session.exit()
		del self.session

	def test_empty_session_wait( self ):
		drmaa_xfail(
				lambda: self.session.wait_for_any(0),
				drmaa.ERRNO_INVALID_JOB
				)

	def test_submit_wait( self ):
		job = sleeper_job( self.session, 2 ).run()
		job.wait_for_state( "running" )
		stat = job.wait()
		self.assertEquals( stat[ "jobid"    ], job.jobid )
		self.assertEquals( stat[ "exited"   ], True )
		self.assertEquals( stat[ "aborted"  ], False )
		self.assertEquals( stat[ "signaled" ], False )
		drmaa_xfail( lambda: job.state(), drmaa.ERRNO_INVALID_JOB )

	def test_submit_in_hold( self ):
		jt = sleeper_job( self.session, 3600 )
		jt.js_state = drmaa.SUBMISSION_STATE_HOLD
		job = jt.run()
		job.wait_for_state( "user_on_hold" )
		job.release()
		job.wait_for_state( "running", "user_on_hold" )
		job.suspend()
		job.wait_for_state( "user_suspended", "running" )
		job.terminate()
		job.wait()
		del job
		del jt



class Test_multi_threaded( unittest.TestCase ):
	def setUp( self ):
		self.n_tester_threads = 5
		self.n_passes = 10
		self.session = drmaa.Session()
		self.mutex = Mutex()
		self.jobs = []
		self.threads = []

	def tearDown( self ):
		del self.threads
		del self.jobs
		self.session.exit()
		del self.session

	def test_submit_terminate_wait( self ):
		self.run_multiple_threads(
				lambda name: Submit_terminate_wait_thread(self, name) )

	def test_control_jobs( self ):
		self.run_multiple_threads(
				lambda name: Control_thread(self, name) )

	def test_stress( self ):
		self.run_multiple_threads(
				lambda name: Stress_thread(self, name) )

	def run_multiple_threads( self, thread_factory ):
		self.threads = [ thread_factory(i) for i in range( self.n_tester_threads ) ]
		# start threads
		for th in self.threads:
			th.start()
		# wait for threads end
		e = None
		for th in self.threads:
			th.join()
			if th.error is not None:
				e = th.error
		if e is not None:
			raise e
		# wait for remaining jobs
		for job in self.jobs:
			try:
				job.terminate()
			except drmaa.drmaa_error:
				pass
		while True:
			job = self.choose_job( 10000 )
			if job is None:
				break
			try: job.wait()
			except drmaa.drmaa_error, e:
				if e.code != drmaa.ERRNO_NO_RUSAGE:
					raise
		# check if session is empty
		drmaa_xfail(
				lambda: self.session.wait( drmaa.JOB_IDS_SESSION_ANY ),
				drmaa.ERRNO_INVALID_JOB
				)

	def n_jobs( self ):
		lock = Lock( self.mutex )
		return len( self.jobs )

	def choose_job( self, threshold ):
		lock = Lock( self.mutex )
		n = len(self.jobs)
		if n == 0:
			return None
		if n * random.random() > threshold:
			return None
		pos = random.randrange( n )
		job = self.jobs[ pos ]
		self.jobs[ pos:pos+1 ] = []
		return job

	def add_job( self, job, state ):
		lock = Lock( self.mutex )
		job.prev_state = drmaa.to_job_state(state)
		assert not job in self.jobs
		self.jobs.append( job )



class Submit_terminate_wait_thread( Tester_thread ):
	def scheadule( self ):
		self.jt  = sleeper_job( self.env.session, 60 )
		for i in range(self.env.n_passes):
			job = self.jt.run()
			job.wait_for_state( "running", "queued_active" )
			time.sleep( 2*random.random() )
			job.terminate()
			job.wait_for_state( "failed", "running" )
			job.wait()
			time.sleep( 1*random.random() )
			del job
		del self.jt



def _actions_factory():
	final_states = {
		drmaa.HOLD      : drmaa.ps_user_on_hold,
		drmaa.RELEASE   : (drmaa.ps_queued_active, drmaa.ps_running),
		drmaa.SUSPEND   : (drmaa.ps_user_suspended, drmaa.ps_running),
		drmaa.RESUME    : drmaa.ps_running,
		drmaa.TERMINATE : drmaa.ps_failed,
	}
	def control( job, action ):
		final_state = final_states[ action ]
		try:
			job.control( action )
		except drmaa.drmaa_error, e:
			state = job.state()
			assert drmaa.ps_running in drmaa.to_job_state_list(final_state) \
				and state == drmaa.ps_running
			return state
		return job.wait_for_state( final_state, job.prev_state )
	hold      = lambda job: control( job, drmaa.HOLD      )
	release   = lambda job: control( job, drmaa.RELEASE   )
	suspend   = lambda job: control( job, drmaa.SUSPEND   )
	resume    = lambda job: control( job, drmaa.RESUME    )
	terminate = lambda job: control( job, drmaa.TERMINATE )
	def dispose( job ):
		job.wait(0)
		return None
	return {
		drmaa.ps_undetermined           :  [],
		drmaa.ps_queued_active          :  [ hold, terminate ],
		drmaa.ps_system_on_hold         :  [ terminate ],
		drmaa.ps_user_on_hold           :  [ release, terminate ],
		drmaa.ps_user_system_on_hold    :  [ release, terminate ],
		drmaa.ps_running                :  [ suspend, terminate ],
		drmaa.ps_system_suspended       :  [ terminate ],
		drmaa.ps_user_suspended         :  [ resume, terminate ],
		drmaa.ps_user_system_suspended  :  [ resume, terminate ],
		drmaa.ps_done                   :  [ dispose ],
		drmaa.ps_failed                 :  [ dispose ],
	}


class Control_thread( Tester_thread ):
	actions = _actions_factory()

	def scheadule( self ):
		for i in range(self.env.n_passes):
			job = self.env.choose_job(2)
			if job is None:
				self.submit_job()
			else:
				self.change_job_state( job )
			time.sleep( 1 )

	def submit_job( self ):
		self.jt  = sleeper_job( self.env.session, 30 )
		if random.random() > 0.3:
			state = drmaa.ps_queued_active
		else:
			self.jt.js_state = drmaa.SUBMISSION_STATE_HOLD
			state = drmaa.ps_user_on_hold
		job = self.jt.run()
		self.env.add_job( job, state )

	def change_job_state( self, job ):
		state = job.state()
		assert job.prev_state.order() <= state.order(), "%s after %s ?!" % (state, job.prev_state)
		possibilities = self.actions[ state ]
		if len(possibilities) > 0:
			action = random.choice( possibilities )
			state = action( job )
		if state is not None:
			self.env.add_job( job, state )



class Stress_thread( Tester_thread ):
	def scheadule( self ):
		for i in range(self.env.n_passes):
			try:
				job = self.env.choose_job( 2 )
				if job is None:
					job = sleeper_job( self.env.session, 10 ).run()
					self.env.add_job( job, drmaa.ps_queued_active )
				else:
					try:
						job.terminate()
						job.wait()
					except:
						self.env.add_job( job, job.state() )
						raise
			except:
				#import traceback
				self.error = Exception_chain()
				#print "Stress_thread: exception catched:"
				#traceback.print_exc()


class DRMAA_C_testsuite( unittest.TestCase ):
	def __init__( self, *args, **kwargs ):
		unittest.TestCase.__init__( self, *args, **kwargs )
		self.basedir = osp.dirname( osp.abspath(
					__file__.replace("build/lib/","") ) )
		self.testsuite = osp.join( self.basedir, "c_testsuite" )
		#print "!!!", self.testsuite
		for filename in ["sleeper_job", "exit_job", "suicide_job"]:
			setattr( self, filename, osp.join( self.basedir, "jobs", filename ) )
		self.__runner = subprocess.check_call

	def _run( self, *args ):
		self.__runner( [self.testsuite]+list(args) )
		#subprocess.check_call( [self.testsuite]+list(args) )
	
	def run_with_debugger( self, options, result ):
		self.__runner = lambda argv: \
				testing.debug.gdb( argv, self, result, options, with_libtool=True )
		self.run( result )

	def test_st_mult_init( self ):
		self._run( "ST_MULT_INIT" )
	def test_st_mult_exit( self ):
		self._run( "ST_MULT_EXIT" )
	def test_st_supported_attr( self ):
		self._run( "ST_SUPPORTED_ATTR" )
	def test_st_supported_vattr( self ):
		self._run( "ST_SUPPORTED_VATTR" )
	def test_st_version( self ):
		self._run( "ST_VERSION" )
	def test_st_drm_system( self ):
		self._run( "ST_DRM_SYSTEM" )
	def test_st_drmaa_impl( self ):
		self._run( "ST_DRMAA_IMPL" )
	def test_st_contact( self ):
		self._run( "ST_CONTACT" )
	def test_st_empty_session_wait( self ):
		self._run( "ST_EMPTY_SESSION_WAIT" )
	def test_st_empty_session_synchronize_dispose( self ):
		self._run( "ST_EMPTY_SESSION_SYNCHRONIZE_DISPOSE" )
	def test_st_empty_session_synchronize_nodispose( self ):
		self._run( "ST_EMPTY_SESSION_SYNCHRONIZE_NODISPOSE" )
	def test_st_empty_session_control( self ):
		self._run( "ST_EMPTY_SESSION_CONTROL", "DRMAA_CONTROL_SUSPEND" )
		self._run( "ST_EMPTY_SESSION_CONTROL", "DRMAA_CONTROL_RESUME" )
		self._run( "ST_EMPTY_SESSION_CONTROL", "DRMAA_CONTROL_HOLD" )
		self._run( "ST_EMPTY_SESSION_CONTROL", "DRMAA_CONTROL_RELEASE" )
		self._run( "ST_EMPTY_SESSION_CONTROL", "DRMAA_CONTROL_TERMINATE" )
	def test_st_submit_wait( self ):
		self._run( "ST_SUBMIT_WAIT", self.sleeper_job )
	def test_st_bulk_submit_wait( self ):
		self._run( "ST_BULK_SUBMIT_WAIT", self.sleeper_job )
	def test_st_bulk_singlesubmit_wait_individual( self ):
		self._run( "ST_BULK_SINGLESUBMIT_WAIT_INDIVIDUAL", self.sleeper_job )
	def test_st_submitmixture_sync_all_dispose( self ):
		self._run( "ST_SUBMITMIXTURE_SYNC_ALL_DISPOSE", self.sleeper_job )
	def test_st_submitmixture_sync_all_nodispose( self ):
		self._run( "ST_SUBMITMIXTURE_SYNC_ALL_NODISPOSE", self.sleeper_job )
	def test_st_submitmixture_sync_allids_dispose( self ):
		self._run( "ST_SUBMITMIXTURE_SYNC_ALLIDS_DISPOSE", self.sleeper_job )
	def test_st_submitmixture_sync_allids_nodispose( self ):
		self._run( "ST_SUBMITMIXTURE_SYNC_ALLIDS_NODISPOSE", self.sleeper_job )
	def test_st_exit_status( self ):
		self._run( "ST_EXIT_STATUS", self.exit_job )
	def test_st_submit_kill_sig( self ):
		self._run( "ST_SUBMIT_KILL_SIG", self.suicide_job )
	def test_st_input_file_failure( self ):
		self._run( "ST_INPUT_FILE_FAILURE", self.sleeper_job )
	def test_st_output_file_failure( self ):
		self._run( "ST_OUTPUT_FILE_FAILURE", self.sleeper_job )
	def test_st_error_file_failure( self ):
		self._run( "ST_ERROR_FILE_FAILURE", self.sleeper_job )
	def test_st_submit_in_hold_release( self ):
		self._run( "ST_SUBMIT_IN_HOLD_RELEASE", self.sleeper_job )
	def test_st_submit_in_hold_delete( self ):
		self._run( "ST_SUBMIT_IN_HOLD_DELETE", self.sleeper_job )
	def test_st_bulk_submit_in_hold_session_release( self ):
		self._run( "ST_BULK_SUBMIT_IN_HOLD_SESSION_RELEASE", self.sleeper_job )
	def test_st_bulk_submit_in_hold_single_release( self ):
		self._run( "ST_BULK_SUBMIT_IN_HOLD_SINGLE_RELEASE", self.sleeper_job )
	def test_st_bulk_submit_in_hold_session_delete( self ):
		self._run( "ST_BULK_SUBMIT_IN_HOLD_SESSION_DELETE", self.sleeper_job )
	def test_st_bulk_submit_in_hold_single_delete( self ):
		self._run( "ST_BULK_SUBMIT_IN_HOLD_SINGLE_DELETE", self.sleeper_job )
	def test_st_submit_polling_wait_timeout( self ):
		self._run( "ST_SUBMIT_POLLING_WAIT_TIMEOUT", self.sleeper_job )
	def test_st_submit_polling_wait_zerotimeout( self ):
		self._run( "ST_SUBMIT_POLLING_WAIT_ZEROTIMEOUT", self.sleeper_job )
	def test_st_submit_polling_synchronize_timeout( self ):
		self._run( "ST_SUBMIT_POLLING_SYNCHRONIZE_TIMEOUT", self.sleeper_job )
	def test_st_submit_polling_synchronize_zerotimeout( self ):
		self._run( "ST_SUBMIT_POLLING_SYNCHRONIZE_ZEROTIMEOUT", self.sleeper_job )
	def test_st_attribute_change( self ):
		self._run( "ST_ATTRIBUTE_CHANGE" )
	def test_st_submit_suspend_resume_wait( self ):
		self._run( "ST_SUBMIT_SUSPEND_RESUME_WAIT", self.sleeper_job )
	def test_st_usage_check( self ):
		self._run( "ST_USAGE_CHECK", self.sleeper_job )
	def test_mt_submit_wait( self ):
		self._run( "MT_SUBMIT_WAIT", self.sleeper_job )
	def test_mt_submit_before_init_wait( self ):
		self._run( "MT_SUBMIT_BEFORE_INIT_WAIT", self.sleeper_job )
	def test_mt_exit_during_submit( self ):
		self._run( "MT_EXIT_DURING_SUBMIT", self.sleeper_job )
	def test_mt_submit_mt_wait( self ):
		self._run( "MT_SUBMIT_MT_WAIT", self.sleeper_job )
	def test_mt_exit_during_submit_or_wait( self ):
		self._run( "MT_EXIT_DURING_SUBMIT_OR_WAIT", self.sleeper_job )


#if __name__ == '__main__':
#	import testing
#	from testing import utils
#	utils.unbuffered_stdout()
#	testing.main(
#			testsuite = testsuite(),
#			prog = "drmaa_testsuite.py",
#			description = "DRMAA testsuite",
#			)
