# $Id: core.py 529 2007-12-20 21:04:22Z lukasz $
import unittest
import optparse
import os
import os.path as osp
import datetime
import sys
import types

from testing import term


def main(
		testsuite,
		prog = None, # program name
		argv = None,
		description = "Testsuite", # testsuite description
		module = "__main__", # module to load tests from (if not given)
		):
	if argv is None:
		argv = sys.argv
	if prog is None:
		prog = osp.basename( argv[0] )

	parser = optparse.OptionParser(
			prog = prog, \
			description = description, \
			usage = "%prog [--gui|--text] [options] [test-names]", \
			conflict_handler = "resolve", \
			)
	parser.add_option( "--gui", help="use gui",
			action="store_const", dest="ui", const="gui" )
	parser.add_option( "--text", help="present all result on console",
			action="store_const", dest="ui", const="text" )
	parser.add_option( "--gdb", help="run tests with gdb",
			action="store_const", dest="ui", const="gdb" )
	parser.add_option( "--gdb-init", "-x",
			help="run gdb commands from given file before running each test" )
	parser.add_option( "--logdir", "-l", dest="log_dir",
			help="write 2 log files (stdout,stderr) per each test into given directory" )
	parser.add_option( "--pretend", "-p", action="store_true",
			help="print tests which would have been run; do not run anything" )
	parser.add_option( "--verbose", "-v", action="count", \
			help="be (more) verbose" )
	parser.add_option( "--quiet", "-q", action="count", \
			help="be (more) quiet" )
	parser.add_option( \
			"--progress", type="choice", \
			choices=("yes","no","auto"), default="auto", \
			help="Controls whether to set terminal title with testsuite progress." \
			+"	PROGRESS can be one of: `yes', `no' or `auto' (default).", \
			)
	parser.add_option( "--help", "-h", action="help",
			help="show this help message and exit" )
	parser.set_defaults( verbose=0, quiet=0, ui="text", pretend=False )
	(options,args) = parser.parse_args( args=argv[1:] )

	options.progress = dict(
			yes=lambda:True,
			no=lambda:False, 
			auto=term.deduce_term_title_support,
			)[ options.progress ]()
	options.verbosity = 1 + options.verbose - options.quiet

	def load_module( module ):
		if type(module) == type(''):
			res = __import__( module )
			for part in module.split('.')[1:]:
				res = getattr( res, part )
		else:
			res = module
		return res
	module = load_module( module )

	loader = TestLoader()
	if len(args) > 0:
		testsuite = loader.loadTestsFromNames( args )
	if testsuite is None:
		testsuite = loader.loadTestsFromModule( module )

	if options.pretend:
		if options.verbosity >= 1:
			print >>sys.stderr, "Would run following tests:"
		for test in iterate_tests(testsuite):
			print test_name(test)
	else:
		result = runner( options.ui )( options ).run( testsuite )
		if result.wasSuccessful():
			sys.exit( 0 )
		else:
			sys.exit( 1 )


def compose_options( options ):
	if options.ui != "gdb":
		yield "--"+options.ui
	if options.verbose > 0:
		yield "-" + "v"*options.verbose
	if options.quiet > 0:
		yield "-" + "q"*options.verbose
	if options.log_dir:
		yield "--logdir=%s" % options.log_dir
	if options.pretend:
		yield "-p"


def runner( runner_name ):
	def text():
		import tui
		return tui.tui_test_runner
	def gui():
		raise NotImplementedError, "gui is not yet supported"
		#import gui
		#return gui.gtk_test_runner
	def gdb():
		import debug
		return debug.gdb_test_runner
	return dict( text=text, gui=gui, gdb=gdb )[ runner_name ]()


#def call_base_before( method ):
#	def wrapper( self, *args, **kwargs ):
#		base_method = getattr( self.__class__.__base__, method.func_name )
#		base_method( self, *args, **kwargs )
#		return method( self, *args, **kwargs )
#	return wrapper


def test_name( test ):
	return "%s.%s.%s" % ( \
			test.__class__.__module__, \
			test.__class__.__name__, \
			test._testMethodName, \
			)


def before_hook( hook ):
	def decorator( f ):
		def wrapper( *args, **kwargs ):
			hook( *args, **kwargs )
			return f( *args, **kwargs )
		return wrapper
	return decorator


def count_tests( test ):
	if hasattr( test, "_tests" ):
		return sum( map( count_tests, test._tests ) )
	else:
		return 1


def iterate_tests( test ):
	if hasattr( test, "_tests" ):
		return sum( map( iterate_tests, test._tests ), [] )
	else:
		return [test]


class File_redirect( object ):
	def __init__( self, fd, filename ):
		self.fd = fd
		self.backup_fd = None
		self.log_fd = None
		try:
			self.backup_fd = os.dup( self.fd )
			#print >>sys.stderr, "backup_fd:", self.backup_fd
			os.close( self.fd )
			self.log_fd = os.open( filename, os.O_CREAT|os.O_WRONLY, 0600 )
			if self.log_fd != self.fd:
				os.dup2( self.log_fd, self.fd )
		except:
			self.close()
			raise
	
	def close( self ):
		if self.backup_fd is not None:
			os.dup2( self.backup_fd, self.fd )
			os.close( self.backup_fd )
			self.backup_fd = None
		if self.log_fd is not None and self.log_fd != self.fd:
			os.close( self.log_fd )
			self.log_fd = None
		
	def __del__( self ):
		self.close()



class Test_runner( object ):
	def __init__( self, test_result_factory ):
		self.test_result_factory = test_result_factory

	def run( self, testsuite ):
		result = self.create_result( testsuite )
		result.beforeTestsuite( testsuite )
		testsuite( result )
		result.afterTestsuite( testsuite )
		return result

	def create_result( self, testsuite ):
		return self.test_result_factory( testsuite )



class Test_info( object ):
	def __init__( self, test, **kwargs ):
		self.test = test
		self.start_time = None
		self.end_time = None
		self.update( **kwargs )
	
	def update( self, **kwargs ):
		for name, value in kwargs.iteritems():
			setattr( self, name, value )

	def time_spend( self ):
		if self.start_time is None:
			return None
		elif self.end_time is None:
			return datetime.datetime.now() - self.start_time
		else:
			return self.end_time - self.start_time

	def __str__( self ):
		return test_name( self.test )



class Test_result_base( unittest.TestResult ):
	def __init__( self, testsuite ):
		unittest.TestResult.__init__( self )
		self.test_running = False
		self.all_tests = count_tests( testsuite )
		self.tests_ran = 0
		self.tests_passed = 0
		self.start_time = None
		self.end_time = None
		self.last_test_finished_at = None
		self.tests = []
	
	def beforeTestsuite( self, testsuite ):
		self.start_time = datetime.datetime.now()
		self.last_test_finished_at = self.start_time

	def afterTestsuite( self, testsuite ):
		self.end_time = datetime.datetime.now()

	def startTest( self, test ):
		unittest.TestResult.startTest( self, test )
		self.test_running = True
		self.tests.append(
				Test_info( test,
					no = self.status()["test_no"],
					start_time = datetime.datetime.now(),
					end_time = None,
					result = None,
					)
				)
	
	def afterTest( self, test, err ):
		self.tests_ran += 1
	
	def addResult( self, test, result ):
		pass

	def __after_test( self, test, err=None ):
		if self.test_running:
			self.test_running = False
			self.afterTest( test, err )

	@before_hook( __after_test )
	def addSuccess( self, test ):
		unittest.TestResult.addSuccess( self, test )
		self.tests_passed += 1
		self.tests[-1].update( result=("success", None) )
		self.addResult( test, "success" )

	@before_hook( __after_test )
	def addError( self, test, err ):
		unittest.TestResult.addError( self, test, err )
		self.tests[-1].result=("error", self.errors[-1][1])
		self.addResult( test, "error" )

	@before_hook( __after_test )
	def addFailure( self, test, err ):
		unittest.TestResult.addFailure( self, test, err )
		self.tests[-1].result=("failure", self.failures[-1][1])
		self.addResult( test, "failure" )

	@before_hook( __after_test )
	def stopTest( self, test ):
		unittest.TestResult.stopTest( self, test )
		self.tests[-1].end_time = datetime.datetime.now()
		self.last_test_finished_at = datetime.datetime.now()

	def status( self ):
		if self.all_tests > 0:
			percent_done = 100. * self.tests_ran / self.all_tests
		else:
			percent_done = 0
		if self.tests_ran > 0:
			success_rate = 100. * self.tests_passed / self.tests_ran
		else:
			success_rate = 0
		if self.start_time is not None:
			time_spend = self.last_test_finished_at - self.start_time
		else:
			time_spend = None
		return dict(
				all_tests = self.all_tests,
				tests_ran = self.tests_ran,
				percent_done = percent_done,
				tests_passed = self.tests_passed,
				success_rate = success_rate,
				errors = len( self.errors ),
				failures = len( self.failures ),
				test_no = (None, self.tests_ran+1)[ self.tests_ran < self.all_tests ],
				start_time = self.start_time,
				end_time = self.end_time,
				time_spend = time_spend,
				)


class TestLoader( object ):
	"""This class is responsible for loading tests according to various
	criteria and returning them wrapped in a Test
	"""
	testMethodPrefix = 'test'
	sortTestMethodsUsing = cmp
	suiteClass = unittest.TestSuite

	def loadTestsFromTestCase(self, testCaseClass):
		"""Return a suite of all tests cases contained in testCaseClass"""
		if issubclass(testCaseClass, unittest.TestSuite):
			raise TypeError("Test cases should not be derived from TestSuite. Maybe you meant to derive from TestCase?")
		testCaseNames = self.getTestCaseNames(testCaseClass)
		if not testCaseNames and hasattr(testCaseClass, 'runTest'):
			testCaseNames = ['runTest']
		return self.suiteClass(map(testCaseClass, testCaseNames))

	def loadTestsFromModule(self, module):
		"""Return a suite of all tests cases contained in the given module"""
		tests = []
		for name in dir(module):
			obj = getattr(module, name)
			if (isinstance(obj, (type, types.ClassType)) and
				issubclass(obj, unittest.TestCase)):
				tests.append(self.loadTestsFromTestCase(obj))
		return self.suiteClass(tests)

	def loadTestsFromName(self, name, module=None):
		"""Return a suite of all tests cases given a string specifier.

		The name may resolve either to a module, a test case class, a
		test method within a test case class, or a callable object which
		returns a TestCase or TestSuite instance.

		The method optionally resolves the names relative to a given module.
		"""
		parts = name.split('.')
		if module is None:
			parts_copy = parts[:]
			while parts_copy:
				try:
					module = __import__('.'.join(parts_copy))
					break
				except ImportError:
					del parts_copy[-1]
					if not parts_copy: raise
			parts = parts[1:]
		obj = module
		parent = None
		name = None
		for part in parts:
			parent, obj, name = obj, getattr(obj, part), part

		if type(obj) == types.ModuleType:
			return self.loadTestsFromModule(obj)
		elif (isinstance(obj, (type, types.ClassType)) and
			  issubclass(obj, TestCase)):
			return self.loadTestsFromTestCase(obj)
		elif type(obj) == types.UnboundMethodType:
			return parent( name )
		elif isinstance(obj, unittest.TestSuite):
			return obj
		elif callable(obj):
			test = obj()
			if not isinstance(test, (unittest.TestCase, unittest.TestSuite)):
				raise ValueError, \
					  "calling %s returned %s, not a test" % (obj,test)
			return test
		else:
			raise ValueError, "don't know how to make test from: %s" % obj

	def loadTestsFromNames(self, names, module=None):
		"""Return a suite of all tests cases found using the given sequence
		of string specifiers. See 'loadTestsFromName()'.
		"""
		suites = [self.loadTestsFromName(name, module) for name in names]
		return self.suiteClass(suites)

	def getTestCaseNames(self, testCaseClass):
		"""Return a sorted sequence of method names found within testCaseClass
		"""
		def isTestMethod(attrname, testCaseClass=testCaseClass, prefix=self.testMethodPrefix):
			return attrname.startswith(prefix) and callable(getattr(testCaseClass, attrname))
		testFnNames = filter(isTestMethod, dir(testCaseClass))
		for baseclass in testCaseClass.__bases__:
			for testFnName in self.getTestCaseNames(baseclass):
				if testFnName not in testFnNames:  # handle overridden methods
					testFnNames.append(testFnName)
		if self.sortTestMethodsUsing:
			testFnNames.sort(self.sortTestMethodsUsing)
		return testFnNames

